/*
 * ext2_fs.c
 *
 * Copyright 2003, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 *
 * EXT2ե륷ƥɥ饤С
 */


#include"types.h"
#include"lib.h"
#include"mm.h"
#include"fs.h"
#include"device.h"
#include"proc.h"
#include"time.h"
#include"errno.h"
#include"test.h"


enum{
	BYTE_BIT=8,		/* 1ХȤΥӥåȿ */
	INT_LOG=2,		/* intΥХȿΣ򺬤Ȥп */
};


/***********************************************************************
 *
 * < Common class >
 *
 ***********************************************************************/

enum{
	EXT2_BLOCK_BASE=1024,			/* ֥å׻뤿Υ١Хȿ */
	EXT2_BLOCK_BASE_LOG=10,			/* ١֥åΣ򺬤Ȥп */

	EXT2_SUPER_BLK_POSITION=1024,	/* ѡ֥åΰ֡ƬΥХȿˡ */

	EXT2_GROUP_DESC_BLK=1,	/* group descriptor block number. */

	/* File system states at "s_state". */
	EXT2_VALID_FS=0x0001,			/* Unmounted cleanly */
	EXT2_ERROR_FS=0x0002,			/* Errors detected */

	/* Maximal mount counts. */
	EXT2_DFL_MAX_MNT_COUNT=20,
	EXT2_DFL_CHECKINTERVAL=0,

	/* Magic number at "s_magic". */
	EXT2_SUPER_MAGIC=0xEF53,

	/* Behaviour when detecting errors. */
	EXT2_ERRORS_CONTINUE=1,			/* continue as if nothing happened. */
	EXT2_ERRORS_RO=2,				/* remount read-only. */
	EXT2_ERRORS_PANIC=3,			/* cause a kernel panic. */
	EXT2_ERRORS_DEFAULT=EXT2_ERRORS_CONTINUE,

	/* revision level value. */
	EXT2_GOOD_OLD_REV=0,			/* original format. */
	EXT2_DYNAMIC_REV=1,				/* V2 format with dynamic inode sizes. */

	/* EXT2_*_INO values at "s_first_ino". */
	EXT2_BAD_INO=0x01, 				/* bad blocks inode */
	EXT2_ROOT_INO=0x02, 			/* root directory inode */
	EXT2_ACL_IDX_INO=0x03, 			/* ACL index inode (deprecated?) */
	EXT2_ACL_DATA_INO=0x04, 		/* ACL data inode (deprecated?) */
	EXT2_BOOT_LOADER_INO=0x05,		/* boot loader inode */
	EXT2_UNDEL_DIR_INO=0x06,		/* undelete directory inode */
};

/* super block. */
typedef struct{
	uint s_inodes_count;			/* 00,total number of inodes. */
	uint s_blocks_count;			/* 04,total number of blocks. */
	uint s_r_blocks_count;			/* 08,total number of blocks reserved for the usage of the super user. */
	uint s_free_blocks_count;		/* 0c,total number of free blocks. */
	uint s_free_inodes_count;		/* 10,total number of free inodes. */
	uint s_first_data_block;		/* 14,the block containing the superblock structure.block size larger than 1KB=0,block size of 1KB=1. */
	uint s_log_block_size;			/* 18,number of bits to shift left the value 1024 for computing block size. */
	int s_log_frag_size;			/* 1c,number of bits to shift the value 1024 for computing fragment size. */
	uint s_blocks_per_group;		/* 20,total number of blocks per group. */
	uint s_frags_per_group;			/* 24,total number of fragments per group. */
	uint s_inodes_per_group;		/* 28,total number of inodes per group. */
	uint s_mtime;					/* 2c,Unix time,last time the file system was mounted. */
	uint s_wtime;					/* 30,Unix time,last write access to the file system. */
	ushort s_mnt_count;				/* 34,how many time the file system was mounted. */
	short s_max_mnt_count;			/* 36,the maximum number of mount times. */
	ushort s_magic;					/* 38,magic number 0xEF53. */
	ushort s_state;					/* 3a,file system state. */
	ushort s_errors;				/* 3c,behaviour when an error is detected. */
	ushort s_minor_rev_level;		/* 3e,minor revision level. */
	uint s_lastcheck;				/* 40,Unix time,last file system check. */
	uint s_checkinterval;			/* 44,between file system checks. */
	uint s_creator_os;				/* 48,os that created the file system. */
	uint s_rev_level;				/* 4c,revision level value. */
	ushort s_def_resuid;			/* 4e,default user id for reserved blocks. */
	ushort s_def_resgid;			/* 50,default group id for reserved blocks. */

	/* EXT2_DYNAMIC_REV Specific */
	uint s_first_ino;				/* 54,index to the first inode useable for standard files. */
	ushort s_inode_size;			/* 58,size of the inode structure. */
	ushort s_block_group_nr;		/* 5a,block group number hosting this superblock. */
	uint s_feature_compat;			/* 5c,bitmask of compatible features. */
	uint s_feature_incompat;		/* 60,bitmask of incompatible features. */
	uint s_feature_ro_compat;		/* 64,bitmask of read-only features. */
	uchar s_uuid[16];				/* 68,volume id. */
	char s_volume_name[16];			/* 78,volume name,mostly unusued. */
	char s_last_mounted[64];		/* 88,last mounted directory path,not normally used. */
	uint s_algo_bitmap;				/* compression algorithms. */

	/* Performance Hints */
	uchar s_prealloc_blocks;		/* nmber of blocks to try to preallocate. */
	uchar s_prealloc_dir_blocks;	/* number to preallocate for dirs. */
	ushort s_padding1;

	/* Journaling Support */
	uchar s_journal_uuid[16];		/*  */
	uint s_journal_inum;			/*  */
	uint s_journal_dev;				/*  */
	uint s_last_orphan;				/*  */

	/* ꡼¸ȼ */
	int blockSize;					/* ֥å */
	int sectorSize;					/*  */
	int sectors;					/* 1֥åΥ */
}SUPER_BLOCK;

/* groupe descriptor. */
typedef struct{
	uint bg_block_bitmap;			/* 0,the first block of the "block bitmap". */
	uint bg_inode_bitmap;			/* 4,the first block of the "inode bitmap". */
	uint bg_inode_table;			/* 8,the first block of the "inode table". */
	ushort bg_free_blocks_count;	/* c,total number of free blocks. */
	ushort bg_free_inodes_count;	/* e,total number of free inodes. */
	ushort bg_used_dirs_count;		/* 10,number of inodes allocated to directories. */
	ushort bg_pad;
	uint bg_reserved[3];
}GROUP_DESC;

/* ӥåȥޥå׹¤Ρ */
typedef struct{
	uchar *bitmap;			/* ӥåȥޥåץХåե */
	ushort currentGd;		/* ߤΥ롼 */
	ushort currentByte;		/* ߤΥӥåȥޥåפΥХȥեåȡ */
}BITMAP;


/* PRIVATE,ӥåȥޥåפǥХ֥ͤåΰ֤롣 */
static const uchar emptyBit[]={
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,6,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,7,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,6,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,
	0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,0,
};

/* ޥȻ롣 */
static SUPER_BLOCK *superBlock[MAX_DEVICE_OPEN];
static GROUP_DESC *groupDesc[MAX_DEVICE_OPEN];
static int groupDescSize[MAX_DEVICE_OPEN];			/* ֥åñ̤Υ */


/*
 * PUBLIC
 * ֥å饻򻻽Ф롣
 * parameters : device inode,block
 * return : secter number
 */
static inline uint blockToSect(int din,uint blk)
{
	return superBlock[din]->blockSize*blk/superBlock[din]->sectorSize;
}


/*
 * PUBLIC
 * ѡ֥åȥ롼ץǥץǥ˽᤹
 * return : 0 or error=-1
 */
static int writeBackSbGd(int din)
{
	SUPER_BLOCK *sb=superBlock[din];
	int sectors=sb->sectors;
	void *buf;


	/* 롼ץǥץǥ˽᤹ */
	write_cache(din,groupDesc[din],groupDescSize[din]*sectors,blockToSect(din,EXT2_GROUP_DESC_BLK+sb->s_first_data_block));
	write_cache(din,groupDesc[din],groupDescSize[din]*sectors,
		blockToSect(din,EXT2_GROUP_DESC_BLK+sb->s_first_data_block+sb->s_blocks_per_group));

	/* ѡ֥åǥ˽᤹ */
	if((buf=kmalloc(sb->blockSize))==NULL)return -1;
	/*read_cache(din,buf,sectors,blockToSect(din,sb->s_first_data_block));*/
	memset(buf,0,sb->blockSize);
	if(sb->s_first_data_block==1)memcpy(buf,sb,sizeof(SUPER_BLOCK));
	else memcpy((char*)buf+EXT2_SUPER_BLK_POSITION,sb,sizeof(SUPER_BLOCK));
	write_cache(din,buf,sectors,blockToSect(din,sb->s_first_data_block));
	if(sb->s_first_data_block==0)
	{
		memset((char*)buf+EXT2_SUPER_BLK_POSITION,0,sizeof(SUPER_BLOCK));
		memcpy(buf,sb,sizeof(SUPER_BLOCK));
	}
	write_cache(din,buf,sectors,blockToSect(din,sb->s_first_data_block+sb->s_blocks_per_group));
	kfree(buf);

	return 0;
}


/***********************************************************************
 *
 * < Block class >
 *
 ***********************************************************************/

/* PUBLIC. */
static BITMAP blockBm[MAX_DEVICE_OPEN];
static BITMAP ReleaseBlkBm[MAX_DEVICE_OPEN];


/*
 * PUBLIC
 * ֥å롣
 * parameters : device inode
 * return : empty block or error=0;
 */
static uint getBlock(int din)
{
	GROUP_DESC *gd=groupDesc[din];
	SUPER_BLOCK *sb=superBlock[din];
	int block;
	int i,last;


	/* ߤΥ롼פ˶֥åʤ */
	if(gd[blockBm[din].currentGd].bg_free_blocks_count==0)
	{
		int sectors=sb->sectors;

		/* ߤΥӥåȥޥåפ᤹ */
		if(write_cache(din,blockBm[din].bitmap,sectors,blockToSect(din,gd[blockBm[din].currentGd].bg_block_bitmap))!=sectors)
			return 0;

		/* ֥åΤ륰롼פõ */
		for(i=blockBm[din].currentGd+1;;++i)
		{
			if(gd[i].bg_block_bitmap==0)i=0;
			if(i==blockBm[din].currentGd)return 0;
			if(gd[i].bg_free_blocks_count!=0)break;
		}
		if(read_cache(din,blockBm[din].bitmap,sectors,blockToSect(din,gd[i].bg_block_bitmap))!=sectors)return 0;
		blockBm[din].currentGd=i;
	}

	/* ӥåȥޥåפθ */
	i=blockBm[din].currentByte;
	last=sb->s_blocks_per_group/BYTE_BIT;
	while(blockBm[din].bitmap[i]==0xff)
		if(++i>=last)i=0;
	blockBm[din].currentByte=i;
	block=i*BYTE_BIT+emptyBit[blockBm[din].bitmap[i]];
	blockBm[din].bitmap[i]|=1<<emptyBit[blockBm[din].bitmap[i]];

	/* ֥åȤ򸺤餹 */
	--gd[blockBm[din].currentGd].bg_free_blocks_count;
	--sb->s_free_blocks_count;

	return block+blockBm[din].currentGd*sb->s_blocks_per_group+sb->s_first_data_block;
}


/*
 * PUBLIC
 * ֥å롣
 * parameters : device inode,block number
 */
static void releaseBlock(int din,uint block)
{
	uchar *bitmap;
	int gd_index;
	SUPER_BLOCK *sb=superBlock[din];
	int sectors=sb->sectors;
	int a;


	/* ֥åӥåȥޥåפθ */
	gd_index=(block-sb->s_first_data_block)/(sb->s_blocks_per_group);
	if(gd_index==blockBm[din].currentGd)bitmap=blockBm[din].bitmap;
	else if(gd_index==ReleaseBlkBm[din].currentGd)bitmap=ReleaseBlkBm[din].bitmap;
	else
	{
		if(write_cache(din,ReleaseBlkBm[din].bitmap,sectors,blockToSect(din,groupDesc[din][ReleaseBlkBm[din].currentGd].bg_block_bitmap))!=
			sectors)return;
		if(read_cache(din,ReleaseBlkBm[din].bitmap,sectors,blockToSect(din,groupDesc[din][gd_index].bg_block_bitmap))!=sectors)
			return;
		ReleaseBlkBm[din].currentGd=gd_index;
		bitmap=ReleaseBlkBm[din].bitmap;
	}

	/* ӥåȥޥåפ˽񤭹ࡣ */
	a=block-gd_index*sb->s_blocks_per_group-sb->s_first_data_block;
	bitmap[a/BYTE_BIT]&=~(1<<a%BYTE_BIT);

	/* ֥åȤ䤹 */
	++groupDesc[din][gd_index].bg_free_blocks_count;
	++sb->s_free_blocks_count;
}


/*
 * PUBLIC
 * ޥȻν
 * parameters : device inode
 * return : groupe descriptor number or error=-1
 */
static int initBlock(int din)
{
	SUPER_BLOCK *sb=superBlock[din];
	GROUP_DESC *gd=groupDesc[din];
	int i;


	/* ֥åӥåȥޥåפɤ߹ࡣ */
	if((blockBm[din].bitmap=(uchar*)kmalloc(sb->blockSize))==NULL)return -1;
	blockBm[din].currentGd=0;
	blockBm[din].currentByte=0;

	for(i=0;gd[i].bg_free_blocks_count==0;++i)
		if(gd[i].bg_block_bitmap==0)return 0;
	if(read_direct(din,blockBm[din].bitmap,sb->sectors,blockToSect(din,gd[i].bg_block_bitmap))!=sb->sectors)return -1;
	blockBm[din].currentGd=i;

	for(i=0;blockBm[din].bitmap[i]==0xff;++i);
	blockBm[din].currentByte=i;

	if((ReleaseBlkBm[din].bitmap=(uchar*)kmalloc(sb->blockSize))==NULL)return -1;
	ReleaseBlkBm[din].currentGd=0;
	ReleaseBlkBm[din].currentByte=0;
	if(read_direct(din,blockBm[din].bitmap,sb->sectors,blockToSect(din,gd[0].bg_block_bitmap))!=sb->sectors)return -1;

	return blockBm[din].currentGd;
}


/*
 * PUBLIC
 * ޥȻν
 * parameters : device inode
 */
static void umountBlock(int din)
{
	int sectors=superBlock[din]->sectors;


	write_cache(din,ReleaseBlkBm[din].bitmap,sectors,blockToSect(din,groupDesc[din][ReleaseBlkBm[din].currentGd].bg_block_bitmap));
	kfree(ReleaseBlkBm[din].bitmap);
	write_cache(din,blockBm[din].bitmap,sectors,blockToSect(din,groupDesc[din][blockBm[din].currentGd].bg_block_bitmap));
	kfree(blockBm[din].bitmap);
}


/***********************************************************************
 *
 * < Inode class >
 *
 ***********************************************************************/

enum{
	/* at i_mode */
	EXT2_S_IFMT=0xF000,		/* format mask */
	EXT2_S_IFSOCK=0xC000,	/* socket */
	EXT2_S_IFLNK=0xA000,	/* symbolic link */
	EXT2_S_IFREG=0x8000,	/* regular file */
	EXT2_S_IFBLK=0x6000,	/* block device */
	EXT2_S_IFDIR=0x4000,	/* directory */
	EXT2_S_IFCHR=0x2000,	/* character device */
	EXT2_S_IFIFO=0x1000,	/* fifo */
	EXT2_S_ISUID=0x0800,	/* SUID */
	EXT2_S_ISGID=0x0400,	/* SGID */
	EXT2_S_ISVTX=0x0200,	/* sticky bit */
	EXT2_S_IRWXU=0x01C0,	/* user access rights mask */
	EXT2_S_IRUSR=0x0100,	/* read */
	EXT2_S_IWUSR=0x0080,	/* write */
	EXT2_S_IXUSR=0x0040,	/* execute */
	EXT2_S_IRWXG=0x0038,	/* group access rights mask */
	EXT2_S_IRGRP=0x0020,	/* read */
	EXT2_S_IWGRP=0x0010,	/* write */
	EXT2_S_IXGRP=0x0008,	/* execute */
	EXT2_S_IRWXO=0x0007,	/* others access rights mask */
	EXT2_S_IROTH=0x0004,	/* read */
	EXT2_S_IWOTH=0x0002,	/* write */
	EXT2_S_IXOTH=0x0001,	/* execute */

	/* behaviour control flags at "i_flags". */
	EXT2_SECRM_FL=0x00000001,		/* secure deletion */
	EXT2_UNRM_FL=0x00000002,		/* record for undelete */
	EXT2_COMPR_FL=0x00000004,		/* compressed file */
	EXT2_SYNC_FL=0x00000008,		/* synchronous updates */
	EXT2_IMMUTABLE_FL=0x00000010,	/* immutable file */
	EXT2_APPEND_FL=0x00000020,		/* append only */
	EXT2_NODUMP_FL=0x00000040,		/* do not dump/delete file */
	EXT2_NOATIME_FL=0x00000080,		/* do not update .i_atime */
	EXT2_DIRTY_FL=0x00000100,		/* dirty (file is in use?) */
	EXT2_COMPRBLK_FL=0x00000200,	/* compressed blocks */
	EXT2_NOCOMPR_FL=0x00000400,		/* access raw compressed data */
	EXT2_ECOMPR_FL=0x00000800,		/* compression error */
	EXT2_BTREE_FL=0x00010000,		/* b-tree format directory */
	EXT2_INDEX_FL=0x00010000,		/* Hash indexed directory */

	EXT2_ROOT_INODE_INDEX=2,		/* 롼inodeΥǥå */

	/* indirect data block index. */
	INODE_DATA_INDIRECT1=12,		/* ܻȥ֥å͡ */
	INODE_DATA_INDIRECT2=13,		/* Ŵܻȥ֥å͡ */
	INODE_DATA_INDIRECT3=14,		/* Ŵܻȥ֥å͡ */

	INODE_DATA_BLOCK_NUM=15,		/* ǡ֥å */

	INODE_BLOCK_SIZE=512,			/* inodei_blocksΥ֥åñ̥Хȿ */
};


typedef struct{
	ushort i_mode;			/* 00,format of the described file and the access rights. */
	ushort i_uid;			/* 02,user id. */
	uint i_size;			/* 04,size of the file in bytes. */
	uint i_atime;			/* 08,Unix time,last time this file was accessed. */
	uint i_ctime;			/* 0c,Unix time,when file was created. */
	uint i_mtime;			/* 10,Unix time,last time this file was modified. */
	uint i_dtime;			/* 14,Unix time,when file was deleted,this value is always 0. */
	ushort i_gid;			/* 18,group id. */
	ushort i_links_count;	/* 1a,how many times this inode is linked.(ǥ쥯ȥ) */
	uint i_blocks;			/* 1c,both currently in used and currently reserved blocks.number of 512 bytes block. */
	uint i_flags;			/* 20,behave when accessing the data for this inode. */
	uint i_number;			/* 24,OS dependent 1,inode number(open) */

	/*
	 * data block number.
	 * 112 direct block number.
	 * 13 indirect block number.
	 * 14 bi-indirect block number.
	 * 15 tri-indirect block number.
	 * Each indirect block array contains as many entries of 32bit block numbers as possible
	 * (to fill one entire block).
	 */
	uint i_block[INODE_DATA_BLOCK_NUM];		/* 28 */
	uint i_generation;		/* 64,file version (used by NFS). */
	uint i_file_acl;		/* 68,block number containing the extended attributes, */
	uint i_dir_acl;			/* 6c,"high size" of the file. */
	uint i_faddr;			/* 70,location of the last file fragment. */
	uint i_osd2[3];			/* 74,OS dependent 2 */
}INODE;

typedef struct{
	const char *path;
	INODE *dirInode;
	INODE *inodeBuf;
}INODE_PARAM;


static INODE *rootInode[MAX_DEVICE_OPEN];		/* PUBLIC root inode. */


/*====================================================================================================*
 *                                       read data                                                    *
 *====================================================================================================*/

typedef int (*READ_DATA)(int,void*,uint*,size_t,size_t,int);


static int readDirectBlock(int,void*,uint*,size_t,size_t,int);
static int readIndirectBlock(int,void*,uint*,size_t,size_t,int);


/* PRIVATE */
static READ_DATA readIndirect[]={readDirectBlock,readIndirectBlock,readIndirectBlock};


/*
 * PRIVATE
 * ǡɹߡ
 * parameters : device inode,buffer,block table,֥åơ֥нߥ,
 *              ֥åơ֥γϥХȿ,node
 * return : ߥ
 */
static int readDirectBlock(int din,void *buf,uint *block_tbl,size_t size,size_t begin,int node)
{
	char *tmp_buf=NULL;
	int block_size=superBlock[din]->blockSize;
	int sectors=superBlock[din]->sectors;
	int rest=size;
	int s,t;


	block_tbl+=begin/block_size;

	/* ϥХ֥ͤåǤʤ硣 */
	if(begin%block_size>0)
	{
		s=begin%block_size;
		t=block_size-s;
		if(rest<t)t=rest;
		if((tmp_buf=(char*)kmalloc(block_size))==NULL)return -1;
		if(read_cache(din,tmp_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		memcpy(buf,tmp_buf+s,t);
		(char*)buf+=t;
		rest-=t;
		++block_tbl;
	}

	for(;rest>=block_size;rest-=block_size)
	{
		if(read_cache(din,buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		(char*)buf+=block_size;
		++block_tbl;
	}

	/* ǸΥХ֥ͤåǤʤ硣 */
	if(rest>0)
	{
		if(tmp_buf==NULL)
			if((tmp_buf=(char*)kmalloc(block_size))==NULL)return size-rest;
		if(read_cache(din,tmp_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		memcpy(buf,tmp_buf,rest);
	}

	kfree(tmp_buf);

	return size;

ERR:
	kfree(tmp_buf);

	return size-rest;
}


/*
 * PRIVATE
 * ɹߡ
 * parameters : device inode,buffer,block table,֥åơ֥нߥ,
 *              ֥åơ֥γϥХȿ,node
 * return : ߥ
 */
static int readIndirectBlock(int din,void *buf,uint *block_tbl,size_t size,size_t begin,int node)
{
	enum{UINT_LOG=2};	/* uintХȥΣ򺬤Ȥп */

	uint *block_buf;
	uint rest=size;
	int sectors=superBlock[din]->sectors;
	int block_size=superBlock[din]->blockSize;
	int idrct_block_log=EXT2_BLOCK_BASE_LOG+superBlock[din]->s_log_block_size-UINT_LOG;
	int idrct_size=block_size<<(idrct_block_log*node);
	int s,t;


	if((block_buf=(uint*)kmalloc(block_size))==NULL)return -ENOMEM;

	block_tbl+=begin/idrct_size;

	/* ϥХ֥ͤåǤʤ硣 */
	{
		s=begin%idrct_size;
		t=idrct_size-s;
		if(rest<t)t=rest;
		if(read_cache(din,block_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		if(readIndirect[node-1](din,buf,block_buf,t,s,node-1)!=t)goto ERR;
		(char*)buf+=t;
		rest-=t;
		++block_tbl;
	}

	for(;rest>=idrct_size;rest-=idrct_size)
	{
		if(read_cache(din,block_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		if(readIndirect[node-1](din,buf,block_buf,idrct_size,0,node-1)!=idrct_size)goto ERR;
		(char*)buf+=idrct_size;
		++block_tbl;
	}

	if(rest>0)
	{
		if(read_cache(din,block_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		if(readIndirect[node-1](din,buf,block_buf,rest,0,node-1)!=rest)goto ERR;
	}

	kfree(block_buf);

	return size;

ERR:
	kfree(block_buf);

	return -1;
}


/*
 * PUBLIC
 * read date.
 * parameters : device inode,inode,buffer,read size,begin byte
 * return : read size or error=-1
 */
static int readData(int din,INODE *inode,void *buf,size_t size,size_t begin)
{
	uint *block_tbl;
	uint rest=size;
	int sectors=superBlock[din]->sectors;
	int block_size=superBlock[din]->blockSize;
	int indirect_blocks=block_size/sizeof(uint);
	int level_size;
	int tmp;


	level_size=block_size*INODE_DATA_INDIRECT1;

	if(begin<level_size)goto DIRECT;
	if((block_tbl=(uint*)kmalloc(block_size))==NULL)return -ENOMEM;
	if(begin<level_size+indirect_blocks*block_size)
	{
		begin-=level_size;
		goto INDIRECT1;
	}
	else if(begin<level_size+(indirect_blocks+1)*indirect_blocks*block_size)
	{
		begin-=level_size+indirect_blocks*block_size;
		goto INDIRECT2;
	}
	else
	{
		begin-=level_size+(indirect_blocks+1)*indirect_blocks*block_size;
		goto INDIRECT3;
	}

DIRECT:		/* ľܥǡ֥å */
	tmp=(rest<=level_size)?rest:level_size;
	if(readDirectBlock(din,buf,inode->i_block,tmp,begin,0)!=tmp)return -1;
	if((rest-=tmp)==0)return size;
	if((block_tbl=(uint*)kmalloc(block_size))==NULL)return -ENOMEM;
	(char*)buf+=tmp;
	begin=0;

INDIRECT1:	/* ܥ֥å */
	level_size=indirect_blocks*block_size;
	tmp=(rest<=level_size)?rest:level_size;
	if(read_cache(din,block_tbl,sectors,blockToSect(din,inode->i_block[INODE_DATA_INDIRECT1]))!=sectors)goto ERR;
	if(readDirectBlock(din,buf,block_tbl,tmp,begin,0)!=tmp)goto ERR;
	if((rest-=tmp)==0)goto END;
	(char*)buf+=tmp;
	begin=0;

INDIRECT2:	/* Ŵܥ֥å */
	level_size=indirect_blocks*indirect_blocks*block_size;
	tmp=(rest<=level_size)?rest:level_size;
	if(read_cache(din,block_tbl,sectors,blockToSect(din,inode->i_block[INODE_DATA_INDIRECT2]))!=sectors)goto ERR;
	if(readIndirectBlock(din,buf,block_tbl,tmp,begin,1)!=tmp)goto ERR;
	if((rest-=tmp)==0)goto END;
	(char*)buf+=tmp;
	begin=0;

INDIRECT3:	/* Ŵܥ֥å */
	if(read_cache(din,block_tbl,sectors,blockToSect(din,inode->i_block[INODE_DATA_INDIRECT3]))!=sectors)goto ERR;
	if(readIndirectBlock(din,buf,block_tbl,rest,begin,2)!=rest)goto ERR;

END:
	kfree(block_tbl);

	return size;

ERR:
	kfree(block_tbl);

	return -1;
}


/*====================================================================================================*
 *                                      write data                                                    *
 *====================================================================================================*/

typedef int (*WRITE_DATA)(int,void*,uint*,size_t,size_t,int);


static int writeDirectBlock(int,void*,uint*,size_t,size_t,int);
static int writeIndirectBlock(int,void*,uint*,size_t,size_t,int);


/* PRIVATE */
static WRITE_DATA writeIndirect[]={writeDirectBlock,writeIndirectBlock,writeIndirectBlock};


/*
 * PRIVATE
 * ǡߡ
 * parameters : device inode,buffer,block table,֥åơ֥нߥ,
 *              ֥åơ֥γϥХȿ,node
 * return : ߥ
 */
static int writeDirectBlock(int din,void *buf,uint *block_tbl,size_t size,size_t begin,int node)
{
	char *tmp_buf=NULL;
	int block_size=superBlock[din]->blockSize;
	int sectors=superBlock[din]->sectors;
	int rest=size;
	int s,t;


	block_tbl+=begin/block_size;

	/*
	 * ƺѥ֥åؤνߡ
	 */
	/* ϥХ֥ͤåǤʤ硣 */
	if(begin%block_size>0)
	{
		s=begin%block_size;
		t=block_size-s;
		if(rest<t)t=rest;
		if((tmp_buf=(char*)kmalloc(block_size))==NULL)return -1;
		if(read_cache(din,tmp_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		memcpy(tmp_buf+s,buf,t);
		if(write_cache(din,tmp_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		(char*)buf+=t;
		rest-=t;
		++block_tbl;
	}

	for(;rest>=block_size;rest-=block_size)
	{
		if(*block_tbl==0)goto NEXT;
		if(write_cache(din,buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		(char*)buf+=block_size;
		++block_tbl;
	}

	/* ǸΥХ֥ͤåǤʤ硣 */
	if(rest>0)
	{
		if(*block_tbl==0)goto NEXT;
		if(tmp_buf==NULL)
			if((tmp_buf=(char*)kmalloc(block_size))==NULL)return size-rest;
		if(read_cache(din,tmp_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		memcpy(tmp_buf,buf,rest);
		if(write_cache(din,tmp_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
	}

	kfree(tmp_buf);

	return size;

NEXT:
	/*
	 * ̤ƥ֥åؤνߡ
	 */
	for(;rest>=block_size;rest-=block_size)
	{
		if((*block_tbl=getBlock(din))==0)goto ERR;
		if(write_cache(din,buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		(char*)buf+=block_size;
		++block_tbl;
	}

	/* ǸΥХ֥ͤåǤʤ硣 */
	if(rest>0)
	{
		if((*block_tbl=getBlock(din))==0)goto ERR;
		if(tmp_buf==NULL)
			if((tmp_buf=(char*)kmalloc(block_size))==NULL)return size-rest;
		memcpy(tmp_buf,buf,rest);
		if(write_cache(din,tmp_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
	}

	kfree(tmp_buf);

	return size;

ERR:
	kfree(tmp_buf);

	return size-rest;
}


/*
 * PRIVATE
 * ܽߡ
 * parameters : device inode,buffer,block table,֥åơ֥нߥ,
 *              ֥åơ֥γϥХȿ,node
 * return : ߥ
 */
static int writeIndirectBlock(int din,void *buf,uint *block_tbl,size_t size,size_t begin,int node)
{
	enum{UINT_LOG=2};	/* uintХȥΣ򺬤Ȥп */

	uint *block_buf;
	uint rest=size;
	int sectors=superBlock[din]->sectors;
	int block_size=superBlock[din]->blockSize;
	int idrct_block_log=EXT2_BLOCK_BASE_LOG+superBlock[din]->s_log_block_size-UINT_LOG;
	int idrct_size=block_size<<(idrct_block_log*node);
	int s,t;


	if((block_buf=(uint*)kmalloc(block_size))==NULL)return -ENOMEM;

	block_tbl+=begin/idrct_size;

	/*
	 * ֥åγ硣
	 */
	{
		s=begin%idrct_size;
		t=idrct_size-s;
		if(rest<t)t=rest;
		if(read_cache(din,block_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		if(writeIndirect[node-1](din,buf,block_buf,t,s,node-1)!=t)goto ERR;
		if(write_cache(din,block_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		(char*)buf+=t;
		rest-=t;
		++block_tbl;
	}

	for(;rest>=idrct_size;rest-=idrct_size)
	{
		if(*block_tbl==0)goto NEXT;
		if(read_cache(din,block_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		if(writeIndirect[node-1](din,buf,block_buf,idrct_size,0,node-1)!=idrct_size)goto ERR;
		if(write_cache(din,block_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		(char*)buf+=idrct_size;
		++block_tbl;
	}

	if(rest>0)
	{
		if(*block_tbl==0)goto NEXT;
		if(read_cache(din,block_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		if(writeIndirect[node-1](din,buf,block_buf,rest,0,node-1)!=rest)goto ERR;
		if(write_cache(din,block_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
	}

	kfree(block_buf);

	return size;

NEXT:
	/*
	 * ֥åγʤ硣
	 */
	for(;rest>=idrct_size;rest-=idrct_size)
	{
		if((*block_tbl=getBlock(din))==0)goto ERR;
		memset(block_buf,0,block_size);
		if(writeIndirect[node-1](din,buf,block_buf,idrct_size,0,node-1)!=idrct_size)goto ERR;
		if(write_cache(din,block_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
		(char*)buf+=idrct_size;
		++block_tbl;
	}

	if(rest>0)
	{
		if((*block_tbl=getBlock(din))==0)goto ERR;
		memset(block_buf,0,block_size);
		if(writeIndirect[node-1](din,buf,block_buf,rest,0,node-1)!=rest)goto ERR;
		if(write_cache(din,block_buf,sectors,blockToSect(din,*block_tbl))!=sectors)goto ERR;
	}

	kfree(block_buf);

	return size;

ERR:
	kfree(block_buf);

	return -1;
}


/*
 * PUBLIC
 * write date.
 * parameters : device inode,inode,buffer,read size,begin byte
 * return : read size or error=-1
 */
static int writeData(int din,INODE *inode,void *buf,size_t size,size_t begin)
{
	uint *block_tbl;
	uint rest=size;
	int sectors=superBlock[din]->sectors;
	int block_size=superBlock[din]->blockSize;
	int indirect_blocks=block_size/sizeof(uint);
	int level_size;
	int tmp;


	level_size=block_size*INODE_DATA_INDIRECT1;

	if(begin<level_size)goto DIRECT;
	if((block_tbl=(uint*)kmalloc(block_size))==NULL)return -ENOMEM;
	if(begin<level_size+indirect_blocks*block_size)
	{
		begin-=level_size;
		goto INDIRECT1;
	}
	else if(begin<level_size+(indirect_blocks+1)*indirect_blocks*block_size)
	{
		begin-=level_size+indirect_blocks*block_size;
		goto INDIRECT2;
	}
	else
	{
		begin-=level_size+(indirect_blocks+1)*indirect_blocks*block_size;
		goto INDIRECT3;
	}

DIRECT:		/* ľܥǡ֥å */
	tmp=(rest<=level_size)?rest:level_size;
	if(writeDirectBlock(din,buf,inode->i_block,tmp,begin,0)!=tmp)return -1;
	if((rest-=tmp)==0)return size;
	if((block_tbl=(uint*)kmalloc(block_size))==NULL)return -ENOMEM;
	(char*)buf+=tmp;
	begin=0;

INDIRECT1:	/* ܥ֥å */
	level_size=indirect_blocks*block_size;
	tmp=(rest<=level_size)?rest:level_size;
	if(inode->i_block[INODE_DATA_INDIRECT1]!=0)
	{
		if(read_cache(din,block_tbl,sectors,blockToSect(din,inode->i_block[INODE_DATA_INDIRECT1]))!=sectors)goto ERR;
	}
	else
	{
		if((inode->i_block[INODE_DATA_INDIRECT1]=getBlock(din))==0)goto ERR;
		memset(block_tbl,0,block_size);
	}
	if(writeDirectBlock(din,buf,block_tbl,tmp,begin,0)!=tmp)goto ERR;

	if(write_cache(din,block_tbl,sectors,blockToSect(din,inode->i_block[INODE_DATA_INDIRECT1]))!=sectors)goto ERR;
	if((rest-=tmp)==0)goto END;
	(char*)buf+=tmp;
	begin=0;

INDIRECT2:	/* Ŵܥ֥å */
	level_size=indirect_blocks*indirect_blocks*block_size;
	tmp=(rest<=level_size)?rest:level_size;
	if(inode->i_block[INODE_DATA_INDIRECT2]!=0)
	{
		if(read_cache(din,block_tbl,sectors,blockToSect(din,inode->i_block[INODE_DATA_INDIRECT2]))!=sectors)goto ERR;
	}
	else
	{
		if((inode->i_block[INODE_DATA_INDIRECT2]=getBlock(din))==0)goto ERR;
		memset(block_tbl,0,block_size);
	}
	if(writeIndirectBlock(din,buf,block_tbl,tmp,begin,1)!=tmp)goto ERR;

	if(write_cache(din,block_tbl,sectors,blockToSect(din,inode->i_block[INODE_DATA_INDIRECT2]))!=sectors)goto ERR;
	if((rest-=tmp)==0)goto END;
	(char*)buf+=tmp;
	begin=0;

INDIRECT3:	/* Ŵܥ֥å */
	if(inode->i_block[INODE_DATA_INDIRECT3]!=0)
	{
		if(read_cache(din,block_tbl,sectors,blockToSect(din,inode->i_block[INODE_DATA_INDIRECT3]))!=sectors)goto ERR;
	}
	else
	{
		if((inode->i_block[INODE_DATA_INDIRECT3]=getBlock(din))==0)goto ERR;
		memset(block_tbl,0,block_size);
	}
	if(writeIndirectBlock(din,buf,block_tbl,rest,begin,2)!=rest)goto ERR;
	if(write_cache(din,block_tbl,sectors,blockToSect(din,inode->i_block[INODE_DATA_INDIRECT3]))!=sectors)goto ERR;

END:
	kfree(block_tbl);
	return size;

ERR:
	kfree(block_tbl);
	return -1;
}


/*====================================================================================================*
 *                                          misc                                                      *
 *====================================================================================================*/

/* PRIVATE. */
static BITMAP inodeBm[MAX_DEVICE_OPEN];


/*
 * PUBLIC
 * inodeֹ롣
 * parameters : device inode
 * return : inode number or error=0
 */
static uint getInodeNumber(int din)
{
	GROUP_DESC *gd=groupDesc[din];
	SUPER_BLOCK *sb=superBlock[din];
	uint inode;
	int i,last;


	/* ߤΥ롼פ˶inodeʤ */
	if(gd[inodeBm[din].currentGd].bg_free_inodes_count==0)
	{
		int sectors=sb->sectors;

		/* ߤΥӥåȥޥåפ᤹ */
		if(write_cache(din,inodeBm[din].bitmap,sectors,blockToSect(din,gd[inodeBm[din].currentGd].bg_inode_bitmap))!=sectors)
			return 0;

		for(i=inodeBm[din].currentGd+1;;++i)
		{
			if(gd[i].bg_inode_bitmap==0)i=0;
			if(i==inodeBm[din].currentGd)return 0;
			if(gd[i].bg_free_inodes_count!=0)break;
		}
		if(read_cache(din,inodeBm[din].bitmap,sectors,blockToSect(din,gd[i].bg_inode_bitmap))!=sectors)return 0;
		inodeBm[din].currentGd=i;
	}

	/* ӥåȥޥåפθ */
	i=inodeBm[din].currentByte;
	last=sb->s_inodes_per_group/BYTE_BIT;
	while(inodeBm[din].bitmap[i]==0xff)
		if(++i>=last)i=0;
	inodeBm[din].currentByte=i;
	inode=i*BYTE_BIT+emptyBit[inodeBm[din].bitmap[i]];
	inodeBm[din].bitmap[i]|=1<<emptyBit[inodeBm[din].bitmap[i]];

	/* ֥åȤ򸺤餹 */
	--gd[inodeBm[din].currentGd].bg_free_inodes_count;
	--sb->s_free_inodes_count;

	return inode+inodeBm[din].currentGd*sb->s_inodes_per_group+1;
}


/*
 * PUBLIC
 * inode롣
 * parameters : device inode,inode number
 */
static void releaseInode(int din,uint inumber)
{
	uchar *bitmap;
	int gd_index;
	SUPER_BLOCK *sb=superBlock[din];
	int sectors=sb->sectors;
	int a;


	/* ֥åӥåȥޥåפθ */
	gd_index=(inumber-1)/(sb->s_inodes_per_group);
	if(gd_index!=inodeBm[din].currentGd)
	{
		if((bitmap=(uchar*)kmalloc(sb->blockSize))==NULL)return;
		if(read_cache(din,bitmap,sectors,blockToSect(din,groupDesc[din][gd_index].bg_inode_bitmap))!=sectors)
			return;
	}
	else bitmap=inodeBm[din].bitmap;

	/* ӥåȥޥåפ˽񤭹ࡣ */
	a=inumber-gd_index*sb->s_inodes_per_group-1;
	bitmap[a/BYTE_BIT]&=~(1<<a%BYTE_BIT);
	if(gd_index!=inodeBm[din].currentGd)
	{
		/* ߤΥ֥å롼פȰפ뤫 */
		if(gd_index==blockBm[din].currentGd)
		{
			uchar *buf;

			buf=inodeBm[din].bitmap;
			inodeBm[din].bitmap=bitmap;
			bitmap=buf;
			inodeBm[din].currentGd=gd_index;
			inodeBm[din].currentByte=a/BYTE_BIT;
		}
		if(write_cache(din,bitmap,sectors,blockToSect(din,groupDesc[din][gd_index].bg_inode_bitmap))!=sectors)return;
		kfree(bitmap);
	}

	/* ֥åȤ䤹 */
	++groupDesc[din][gd_index].bg_free_inodes_count;
	++sb->s_free_inodes_count;
}


/*
 * PUBLIC
 * Υ֥åХåեinodeơ֥ɤ߹ǡinodeɥ쥹֤
 * parameters : device inode,inode number,block buffer
 * return : inode addres or error number;
 */
static inline int readInode(int din,uint inumber,INODE *sector_buf)
{
	int gdesc;
	int index;
	uint sector;


	gdesc=(inumber-1)/superBlock[din]->s_inodes_per_group;
	index=(inumber-1)%superBlock[din]->s_inodes_per_group;
	sector=blockToSect(din,groupDesc[din][gdesc].bg_inode_table)+index*sizeof(INODE)/superBlock[din]->sectorSize;
	if(read_cache(din,sector_buf,1,sector)!=1)return -EIO;

	return (int)sector_buf+index*sizeof(INODE)%superBlock[din]->sectorSize;
}


/*
 * PUBLIC
 * inodeǥ˽񤭹ࡣ
 * parameters : device inode,inode number,block buffer
 * return : 0 or error number
 */
static inline int writeInode(int din,uint inumber,INODE *sector_buf)
{
	int gdesc;
	int index;
	uint sector;


	gdesc=(inumber-1)/superBlock[din]->s_inodes_per_group;
	index=(inumber-1)%superBlock[din]->s_inodes_per_group;
	sector=blockToSect(din,groupDesc[din][gdesc].bg_inode_table)+index*sizeof(INODE)/superBlock[din]->sectorSize;
	if(write_cache(din,sector_buf,1,sector)!=1)return -EIO;

	return 0;
}


/*
 * PUBLIC
 * ޥȻν
 * parameters : device inode,򳫻Ϥ֥å롼ץʥС
 * return : 0 or error=-1
 */
static int initInode(int din,int begin_gr)
{
	void *buf=NULL;
	GROUP_DESC *gd=groupDesc[din];
	SUPER_BLOCK *sb=superBlock[din];
	int i;


	/* inodeӥåȥޥåפɤ߹ࡣ */
	if((inodeBm[din].bitmap=(uchar*)kmalloc(sb->blockSize))==NULL)return -1;
	inodeBm[din].currentGd=begin_gr;
	inodeBm[din].currentByte=0;

	i=begin_gr;
	do
	{
		if(gd[i].bg_free_inodes_count>0)
		{
			if(read_direct(din,inodeBm[din].bitmap,sb->sectors,blockToSect(din,gd[i].bg_inode_bitmap))!=sb->sectors)return -1;
			inodeBm[din].currentGd=i;
			for(i=0;inodeBm[din].bitmap[i]==0xff;++i);
			inodeBm[din].currentByte=i;
			break;
		}
		if(gd[++i].bg_inode_bitmap==0)i=0;
	}while(i!=begin_gr);

	/* 롼inodeФ롣 */
	if((buf=kmalloc(sb->sectorSize))==NULL)return -1;
	if((rootInode[din]=(INODE*)kmalloc(sizeof(INODE)))==NULL)goto ERR;
	if(read_direct(din,buf,1,blockToSect(din,gd[0].bg_inode_table))!=1)goto ERR;
	memcpy(rootInode[din],&((INODE*)buf)[EXT2_ROOT_INODE_INDEX-1],sizeof(INODE));
	kfree(buf);
	rootInode[din]->i_number=EXT2_ROOT_INODE_INDEX;

	return 0;

ERR:
	kfree(rootInode[din]);
	kfree(buf);
	return -1;
}


/*
 * PUBLIC
 * ޥȻν
 * parameters : device inode
 */
static int umountInode(int din)
{
	int rest;
	INODE *sector_buf;


	/* 롼inodeǥ˽᤹ */
	if((sector_buf=(INODE*)kmalloc(superBlock[din]->sectorSize))==NULL)return -ENOMEM;
	if((rest=readInode(din,EXT2_ROOT_INODE_INDEX,sector_buf))<0)return rest;
	memcpy((void*)rest,rootInode[din],sizeof(INODE));
	writeInode(din,EXT2_ROOT_INODE_INDEX,sector_buf);
	kfree(sector_buf);
	rootInode[din]=NULL;

	/* inodeӥåȥޥåפǥ˽᤹ */
	write_cache(din,inodeBm[din].bitmap,superBlock[din]->sectors,blockToSect(din,groupDesc[din][inodeBm[din].currentGd].bg_inode_bitmap));
	kfree(inodeBm[din].bitmap);

	return 0;
}


/*====================================================================================================*
 *                                       release                                                      *
 *====================================================================================================*/

typedef int (*RELEASE_DATA)(int,uint*,int,int);


static int releaseDirect(int,uint*,int,int);
static int releaseIndirect(int,uint*,int,int);
RELEASE_DATA releaseData[]={releaseDirect,releaseIndirect,releaseIndirect,releaseIndirect};


/*
 * PRIVATE
 * ľܥ֥åơ֥Υ֥å롣
 * parameters : device inode,block table,number of blocks,node=0
 * return : 0
 */
static int releaseDirect(int din,uint *table,int blk_num,int node)
{
	int i;


	for(i=0;i<blk_num;++i)releaseBlock(din,table[i]);

	return 0;
}


/*
 * PRIVATE
 * ܥ֥åơ֥Υ֥å롣
 * parameters : device inode,block table,number of blocks,node
 * return : 0 or error number
 */
static int releaseIndirect(int din,uint *table,int blk_num,int node)
{
	uint *buf;
	int sectors=superBlock[din]->sectors;
	int nblk=1<<((EXT2_BLOCK_BASE_LOG+superBlock[din]->s_log_block_size-INT_LOG)*node);
	int rest;
	int i;


	if((buf=(uint*)kmalloc(superBlock[din]->blockSize))==NULL)return -ENOMEM;

	i=0;
	for(;blk_num>nblk;blk_num-=nblk)
	{
		if((rest=read_cache(din,buf,sectors,blockToSect(din,table[i])))<0)goto END;
		if((rest=releaseData[node-1](din,buf,nblk,node-1))<0)goto END;
		releaseBlock(din,table[i++]);
	}
	if((rest=read_cache(din,buf,sectors,blockToSect(din,table[i])))<0)goto END;
	if((rest=releaseData[node-1](din,buf,blk_num,node-1))<0)goto END;
	releaseBlock(din,table[i]);

	rest=0;
END:
	kfree(buf);

	return rest;
}


/*
 * PUBLIC
 * ǡ֥å롣
 * parameters : device inode,inode address
 * return : 0 or error number
 */
static int releaseDataBlock(int din,INODE *inode)
{
	uint *buf;
	int sectors=superBlock[din]->sectors;
	int block_size=superBlock[din]->blockSize;
	int blk_num=ROUNDUP_DIV(inode->i_size,block_size);
	int nblk;
	int rest;
	int blk,i;


	/* ľܥ֥å */
	blk=(blk_num>=INODE_DATA_INDIRECT1)?INODE_DATA_INDIRECT1:blk_num;
	releaseDirect(din,inode->i_block,blk,0);
	for(i=0;i<blk;++i)inode->i_block[i]=0;
	if((blk_num-=blk)==0)goto END;

	/* ܥ֥å */
	if((buf=(uint*)kmalloc(block_size))==NULL)return -ENOMEM;
	blk=nblk=block_size/sizeof(uint);
	for(i=INODE_DATA_INDIRECT1;;++i)
	{
		if((rest=read_cache(din,buf,sectors,blockToSect(din,inode->i_block[i])))<0)goto ERR;
		if((rest=releaseData[i-INODE_DATA_INDIRECT1](din,buf,(blk_num>blk)?blk:blk_num,i-INODE_DATA_INDIRECT1))<0)goto ERR;
		releaseBlock(din,inode->i_block[i]);
		inode->i_block[i]=0;
		if((blk_num-=blk)<=0)break;
		blk*=nblk;
	}
	kfree(buf);
END:
	inode->i_size=0;

	return 0;
ERR:
	kfree(buf);

	return rest;
}


/***********************************************************************
 *
 * < Directory class >
 *
 ***********************************************************************/

enum{
	EXT2_NAME_LEN=255,

	/* directory file types. */
	EXT2_FT_UNKNOWN=0,
	EXT2_FT_REG_FILE,
	EXT2_FT_DIR,
	EXT2_FT_CHRDEV,
	EXT2_FT_BLKDEV,
	EXT2_FT_FIFO,
	EXT2_FT_SOCK,
	EXT2_FT_SYMLINK,
	EXT2_FT_MAX,
};


typedef struct{
	uint inode;				/* inode number. */
	ushort rec_len;			/* directory entry length. */
	uchar name_len;			/* name length. */
	uchar file_type;		/* file type. */
	char name[0];			/* File name. */
}DIR_ENTRY;

typedef struct{
	const char *path;
	DIR_ENTRY *dirBuf;
	uint blkNumber;
}DIR_PARAM;

typedef int (*READ_DIR)(int,uint*,int,DIR_PARAM*,int,int (**READ_DIR)());


#define ROUNDUP4(a) (((a+3)>>2)<<2)		/* 4Ƿ夲 */


/*
 * PRIVATE
 * search entry.
 * parameters : device inode,directory block,directory parameters struct
 * return : directory entry address or failed=NULL
 */
static inline DIR_ENTRY *searchEntry(int din,DIR_PARAM *dparam)
{
	DIR_ENTRY *entry,*last;


	entry=dparam->dirBuf;
	if(entry->inode==EXT2_FT_UNKNOWN)(char*)entry+=entry->rec_len;
	(char*)last=(char*)dparam->dirBuf+superBlock[din]->blockSize;

	for(;entry<last;(char*)entry+=entry->rec_len)
		if(cmpPathLength(entry->name,dparam->path,entry->name_len)==0)return entry;

	return NULL;
}


/*
 * PRIVATE
 * ȥ꡼õ
 * parameters : device inode,directory block,directory parameters struct
 * return : directory entry address or failed=NULL
 */
static inline DIR_ENTRY *searchEntrySpace(int din,DIR_PARAM *dparam)
{
	int add_len,ent_len;
	DIR_ENTRY *entry,*last,*new;


	add_len=strlen(dparam->path)+sizeof(DIR_ENTRY);
	entry=dparam->dirBuf;
	(char*)last=(char*)dparam->dirBuf+superBlock[din]->blockSize;

	if(entry->inode==EXT2_FT_UNKNOWN)
	{
		if(entry->rec_len>=add_len)
		{
			entry->name_len=add_len-sizeof(DIR_ENTRY);
			return entry;
		}
		else (char*)entry+=entry->rec_len;
	}

	for(;entry<last;(char*)entry+=entry->rec_len)
	{
		ent_len=ROUNDUP4(entry->name_len+sizeof(DIR_ENTRY));
		if((entry->rec_len-ent_len)>=add_len)
		{
			new=(DIR_ENTRY*)((char*)entry+ent_len);
			new->rec_len=entry->rec_len-ent_len;		/* ȥ꡼ */
			new->name_len=add_len-sizeof(DIR_ENTRY);	/* ͡ॵ */
			entry->rec_len=ent_len;
			return new;
		}
	}

	return NULL;
}


/*
 * PRIVATE
 * ȥ꡼뤫å롣
 * parameters : device inode,directory block,directory parameters struct
 * return : 0 or ȥ꡼=-1
 */
static inline int checkEntry(int din,DIR_PARAM *dparam)
{
	DIR_ENTRY *entry;


	entry=dparam->dirBuf;

	if(entry->inode==EXT2_FT_UNKNOWN)
	{
		if(entry->rec_len==superBlock[din]->blockSize)return 0;
	}
	else
	{
		if(cmpPathLength(entry->name,".",entry->name_len)==0)
		{
			(char*)entry+=entry->rec_len;
			if(entry->rec_len==superBlock[din]->blockSize-(sizeof(DIR_ENTRY)+sizeof(int)))return 0;
		}
	}

	return -1;
}


/*
 * PRIVATE
 * ȥ꡼롣
 * parameters : delete entry,directory buffer
 */
static inline void delEntry(DIR_ENTRY *entry,DIR_ENTRY *dir_buf)
{
	DIR_ENTRY *p;


	if(entry==dir_buf)entry->inode=EXT2_FT_UNKNOWN;
	else
	{
		for(p=dir_buf;(char*)p+p->rec_len!=(char*)entry;(char*)p+=p->rec_len);
		p->rec_len+=entry->rec_len;
	}
}


/*
 * PRIVATE
 * ľܥ֥åɹߡ
 * ͤͿ줿ХåեΥǥ쥯ȥꥢɥ쥹
 * parameters : device inode,block table,number of block table indexs,directory parameters struct,node=0,READ_DIR struct
 * return : directory entry address or error number
 */
static int readDirectDir(int din,uint *table,int blk_num,DIR_PARAM *dparam,int node,READ_DIR *readDirBlock)
{
	int sectors=superBlock[din]->sectors;
	DIR_ENTRY *entry;
	int i;


	for(i=0;i<blk_num;++i)
	{
		if(read_cache(din,dparam->dirBuf,sectors,blockToSect(din,table[i]))!=sectors)return -EIO;
		if((entry=searchEntry(din,dparam))!=NULL)
		{
			dparam->blkNumber=table[i];
			return (int)entry;
		}
	}

	return -ENOENT;
}


/*
 * PRIVATE
 * ľܥ֥åɹߡ
 * ͤͿ줿ХåեΥǥ쥯ȥꥢɥ쥹
 * parameters : device inode,block table,number of block table indexs,directory parameters struct,node=0,READ_DIR struct
 * return : directory entry address or error number
 */
static int searchDirectDir(int din,uint *table,int blk_num,DIR_PARAM *dparam,int node,READ_DIR *readDirBlock)
{
	int sectors=superBlock[din]->sectors;
	DIR_ENTRY *entry;
	int i;


	for(i=0;i<blk_num;++i)
	{
		if(read_cache(din,dparam->dirBuf,sectors,blockToSect(din,table[i]))!=sectors)return -EIO;
		if((entry=searchEntrySpace(din,dparam))!=NULL)
		{
			dparam->blkNumber=table[i];
			return (int)entry;
		}
	}

	return -ENOENT;
}


/*
 * PRIVATE
 * ľܥ֥åɹߡ
 * parameters : device inode,block table,number of block table indexs,directory parameters struct,node=0,READ_DIR struct
 * return : ʤ=-ENOENT or error number
 */
static int checkDirectDir(int din,uint *table,int blk_num,DIR_PARAM *dparam,int node,READ_DIR *readDirBlock)
{
	int sectors=superBlock[din]->sectors;
	int i;


	for(i=0;i<blk_num;++i)
	{
		if(read_cache(din,dparam->dirBuf,sectors,blockToSect(din,table[i]))!=sectors)return -EIO;
		if(checkEntry(din,dparam)==-1)return -EEXIST;
	}

	return -ENOENT;
}


/*
 * PRIVATE
 * ɹߡ
 * ͤͿ줿ХåեΥǥ쥯ȥꥢɥ쥹
 * parameters : device inode,block table,number of block table indexs,directory parameters struct,node,READ_DIR struct
 * return : directory entry address of error number
 */
static int readIndirectDir(int din,uint *table,int blk_num,DIR_PARAM *dparam,int node,READ_DIR *readDirBlock)
{
	uint *buf;
	int sectors=superBlock[din]->sectors;
	int nblk=1<<((EXT2_BLOCK_BASE_LOG+superBlock[din]->s_log_block_size-INT_LOG)*node);
	int rest;
	int i;


	if((buf=(uint*)kmalloc(superBlock[din]->blockSize))==NULL)return -ENOMEM;

	for(i=0;blk_num>nblk;blk_num-=nblk,++i)
	{
		if((rest=read_cache(din,buf,sectors,blockToSect(din,table[i])))<0)goto END;
		if((rest=readDirBlock[node-1](din,buf,nblk,dparam,node-1,readDirBlock))!=-ENOENT)goto END;
	}
	if((rest=read_cache(din,buf,sectors,blockToSect(din,table[i])))<0)goto END;
	rest=readDirBlock[node-1](din,buf,blk_num,dparam,node-1,readDirBlock);

END:
	kfree(buf);

	return rest;
}


/*
 * PRIVATE
 * read directory.
 * ͤͿ줿ХåեΥǥ쥯ȥꥢɥ쥹
 * parameters : path,device inode,inode,directory parameters struct,READ_DIR struct
 * return : directory address or error number
 */
static int readDir(int din,INODE *inode,DIR_PARAM *dparam,READ_DIR *readDirBlock)
{
	uint *buf;
	int sectors=superBlock[din]->sectors;
	int block_size=superBlock[din]->blockSize;
	int blk_num=ROUNDUP_DIV(inode->i_size,block_size);
	int nblk;
	int rest;
	int blk,i;


	/* ľܥ֥å */
	blk=(blk_num>=INODE_DATA_INDIRECT1)?INODE_DATA_INDIRECT1:blk_num;
	if((rest=readDirBlock[0](din,inode->i_block,blk,dparam,0,readDirBlock))!=-ENOENT)return rest;
	if((blk_num-=blk)==0)return -ENOENT;

	/* ܥ֥å */
	if((buf=(uint*)kmalloc(block_size))==NULL)return -ENOMEM;
	nblk=block_size/sizeof(uint);
	blk=nblk;
	for(i=INODE_DATA_INDIRECT1;;++i)
	{
		if((rest=read_cache(din,buf,sectors,blockToSect(din,inode->i_block[i])))<0)goto END;
		if((rest=readDirBlock[i-INODE_DATA_INDIRECT1](din,buf,(blk_num>blk)?blk:blk_num,dparam,i-INODE_DATA_INDIRECT1,readDirBlock))!=-ENOENT)
			goto END;
		if((blk_num-=blk)<=0)break;
		blk*=nblk;
	}
END:
	kfree(buf);

	return rest;
}


static READ_DIR readDirBlock[]={readDirectDir,readIndirectDir,readIndirectDir,readIndirectDir};
static READ_DIR addDirBlock[]={searchDirectDir,readIndirectDir,readIndirectDir,readIndirectDir};
static READ_DIR checkDirBlock[]={checkDirectDir,readIndirectDir,readIndirectDir,readIndirectDir};


/*
 * PUBLIC
 * search directory.
 * ͤϡͿ줿Хåեinodeɥ쥹
 * parameters : device inode,INODE_PARAM struct
 * return : inode address or error number
 */
static int searchPath(int din,INODE_PARAM *param)
{
	int inode;
	DIR_ENTRY *entry;
	DIR_PARAM dparam;


	if((dparam.dirBuf=(DIR_ENTRY*)kmalloc(superBlock[din]->blockSize))==NULL)return -ENOMEM;
	dparam.path=param->path;

	for(;;)
	{
		/* ǥ쥯ȥꥨȥ꡼õ */
		if(((int)entry=readDir(din,param->dirInode,&dparam,readDirBlock))<0)
		{
			param->path=dparam.path;
			inode=(int)entry;
			break;
		}

		/* inode롣 */
		if((inode=readInode(din,entry->inode,param->inodeBuf))<0)break;
		((INODE*)inode)->i_number=entry->inode;

		dparam.path+=entry->name_len;
		if(*dparam.path++=='\0')break;
		if(entry->file_type!=EXT2_FT_DIR)
		{
			inode=-ENOTDIR;
			break;
		}
		param->dirInode=(INODE*)inode;
	}

	kfree(dparam.dirBuf);

	return inode;
}


/*
 * PUBLIC
 * ǥ쥯ȥꥨȥ꡼ä롣
 * ͤϡͿ줿Хåեinodeɥ쥹
 * parameters : device inode,file type,inode number,INODE_PARAM struct
 * return : 0 or error number
 */
static int addDirEnt(int din,int type,uint inumber,INODE_PARAM *param)
{
	int rest=0;
	int block_size=superBlock[din]->blockSize;
	int sectors=superBlock[din]->sectors;
	DIR_ENTRY *new_ent;
	DIR_PARAM dparam;


	if((dparam.dirBuf=(DIR_ENTRY*)kmalloc(block_size))==NULL)return -ENOMEM;
	dparam.path=param->path;

	/* ǥ쥯ȥ꤫ȥ꡼õ */
	if(((int)new_ent=readDir(din,param->dirInode,&dparam,addDirBlock))<0)
	{
		/* ˥ǥ쥯ȥ˥֥åƤ롣 */
		if(superBlock[din]->s_free_blocks_count==0)rest=-ENOSPC;
		else
		{
			new_ent=dparam.dirBuf;
			new_ent->rec_len=block_size;
			new_ent->name_len=strlen(dparam.path);
			memcpy(new_ent->name,param->path,new_ent->name_len);
			new_ent->inode=inumber;
			new_ent->file_type=type;

			if(writeData(din,param->dirInode,new_ent,block_size,param->dirInode->i_size)==block_size)
			{
				param->dirInode->i_size+=block_size;
				param->dirInode->i_blocks=ROUNDUP_DIV(param->dirInode->i_size,INODE_BLOCK_SIZE);
				param->dirInode->i_atime=param->dirInode->i_mtime=(uint)sys_time();
			}
			else rest=-EIO;
		}
	}
	else
	{
		/* ȥ꡼롣 */
		memcpy(new_ent->name,param->path,new_ent->name_len);
		new_ent->inode=inumber;
		new_ent->file_type=type;
		if(write_cache(din,dparam.dirBuf,sectors,blockToSect(din,dparam.blkNumber))!=sectors)rest=-EIO;
	}

	kfree(dparam.dirBuf);

	return rest;
}


/*
 * PRIVATE
 * ǥ쥯ȥ˥ȥ꡼ĤäƤ뤫ǧ롣
 * parameters : device inode,directory inode
 * return : ȥ꡼ʤ=0 or error number
 */
static int inline checkDir(int din,INODE *inode)
{
	int rest;
	DIR_PARAM param;


	if((param.dirBuf=(DIR_ENTRY*)kmalloc(superBlock[din]->blockSize))==NULL)return -ENOMEM;
	rest=readDir(din,inode,&param,checkDirBlock);
	kfree(param.dirBuf);

	return rest;
}


/*
 * PUBLIC
 * ѥΥեǥ쥯ȥꥨȥ꡼롣
 * parameters : name,device inode,INODE_PARAM struct
 * return inode address or error number
 */
static int delFileEntry(int din,INODE_PARAM *param)
{
	int sectors=superBlock[din]->sectors;
	int rest;
	DIR_ENTRY *entry;
	DIR_PARAM dparam;
	INODE *inode;


	if((dparam.dirBuf=(DIR_ENTRY*)kmalloc(superBlock[din]->blockSize))==NULL)return -ENOMEM;
	dparam.path=param->path;

	for(;;)
	{
		/* ǥ쥯ȥꥨȥ꡼õ */
		if((rest=readDir(din,param->dirInode,&dparam,readDirBlock))<0)goto ERR;
		entry=(DIR_ENTRY*)rest;

		dparam.path+=entry->name_len;

		if(*dparam.path++=='\0')
		{
			if(entry->file_type==EXT2_FT_REG_FILE)
			{
				if(((int)inode=readInode(din,entry->inode,param->inodeBuf))<0)break;
				inode->i_number=entry->inode;
				delEntry(entry,dparam.dirBuf);
				if((rest=write_cache(din,dparam.dirBuf,sectors,blockToSect(din,dparam.blkNumber)))<0)goto ERR;
			}
			else (int)inode=-ENOENT;

			break;
		}

		/* inode롣 */
		if(((int)inode=readInode(din,entry->inode,param->inodeBuf))<0)break;
		inode->i_number=entry->inode;
		param->dirInode=inode;
	}

	kfree(dparam.dirBuf);

	return (int)inode;
ERR:
	kfree(dparam.dirBuf);

	return rest;
}


/*
 * PUBLIC
 * ѥΥǥ쥯ȥꥨȥ꡼롣
 * ͤϡͿ줿Хåեinodeɥ쥹
 * parameters : name,device inode,INODE_PARAM struct
 * return inode address or error number
 */
static int delDirEntry(int din,INODE_PARAM *param)
{
	int sectors=superBlock[din]->sectors;
	int rest;
	DIR_ENTRY *entry;
	DIR_PARAM dparam;
	INODE *inode;
	INODE *parent_inode;
	INODE *parent_ibuf=NULL;


	if((dparam.dirBuf=(DIR_ENTRY*)kmalloc(superBlock[din]->blockSize))==NULL)return -ENOMEM;
	dparam.path=param->path;
	inode=param->dirInode;

	for(;;)
	{
		/* ǥ쥯ȥꥨȥ꡼õ */
		if((rest=readDir(din,param->dirInode,&dparam,readDirBlock))<0)goto ERR;
		entry=(DIR_ENTRY*)rest;

		dparam.path+=entry->name_len;

		if(*dparam.path++=='\0')
		{
			if(entry->file_type==EXT2_FT_DIR)
			{
				parent_inode=inode;
				parent_ibuf=param->inodeBuf;

				if((param->inodeBuf=(INODE*)kmalloc(superBlock[din]->sectorSize))==NULL)
				{
					rest=-ENOMEM;
					goto ERR;
				}
				if((rest=readInode(din,entry->inode,param->inodeBuf))<0)goto ERR;
				inode=(INODE*)rest;
				inode->i_number=entry->inode;
				if((rest=checkDir(din,inode))!=-ENOENT)goto ERR;
				delEntry(entry,dparam.dirBuf);
				if((rest=write_cache(din,dparam.dirBuf,sectors,blockToSect(din,dparam.blkNumber)))<0)goto ERR;
				--parent_inode->i_links_count;
				if((rest=writeInode(din,parent_inode->i_number,parent_ibuf))<0)goto ERR;

				kfree(parent_ibuf);
			}
			else (int)inode=-ENOENT;

			break;
		}

		/* inode롣 */
		if(((int)inode=readInode(din,entry->inode,param->inodeBuf))<0)break;
		inode->i_number=entry->inode;
		param->dirInode=inode;
	}

	kfree(dparam.dirBuf);

	return (int)inode;
ERR:
	kfree(dparam.dirBuf);
	kfree(parent_ibuf);

	return rest;
}


/*
 * PUBLIC
 * ǥ쥯ȥ֥åν
 * parameters : device inode,block number,inode number,parent inode number
 * return : 0 or error number
 */
static int initDirBlock(int din,uint block,uint inumber,uint parent_inumber)
{
	int error=0;
	DIR_ENTRY *dent,*dent2;


	if((dent=(DIR_ENTRY*)kmalloc(superBlock[din]->sectorSize))==NULL)return -ENOMEM;
	memset(dent,0,superBlock[din]->sectorSize);			/* ꥢʤȥǥ쥯ȥǷٹθˤʤ롣 */

	/* "."ǥ쥯ȥɲá */
	dent->inode=inumber;
	dent->rec_len=sizeof(DIR_ENTRY)+sizeof(int);
	dent->name_len=1;
	dent->file_type=EXT2_FT_DIR;
	dent->name[0]='.';

	(char*)dent2=(char*)dent+dent->rec_len;

	/* ".."ǥ쥯ȥɲá */
	dent2->inode=parent_inumber;
	dent2->rec_len=superBlock[din]->blockSize-dent->rec_len;
	dent2->name_len=2;
	dent2->file_type=EXT2_FT_DIR;
	dent2->name[0]='.';
	dent2->name[1]='.';

	if(write_cache(din,dent,1,blockToSect(din,block))!=1)error=-EIO;

	kfree(dent);

	return error;
}


/***********************************************************************
 *
 * < system call >
 *
 ***********************************************************************/

/*
 * return : root inode address in memory.
 */
static uint mount(int din)
{
	void *buf;
	DEV_STAT dev_stat;
	int sectors;
	int gd;
	int a;


	/* Ǥ˥ޥȤƤ뤫ǧ */
	if(superBlock[din]!=NULL)return -1;

	/* ѡ֥åɤ߹ࡣ */
	if(get_dev_stat(din,&dev_stat)==-1)return -1;
	if((buf=kmalloc(PAGE_SIZE))==NULL)return -1;
	if(read_direct(din,buf,PAGE_SIZE/dev_stat.sect_size,0)!=PAGE_SIZE/dev_stat.sect_size)
		{kfree(buf);return -1;}
	if((superBlock[din]=(SUPER_BLOCK*)kmalloc(sizeof(SUPER_BLOCK)))==NULL)
		{kfree(buf);return -1;}
	memcpy(superBlock[din],(char*)buf+EXT2_SUPER_BLK_POSITION,sizeof(SUPER_BLOCK));
	kfree(buf);

	/* ޥåʥСγǧ */
	if(superBlock[din]->s_magic!=EXT2_SUPER_MAGIC)goto ERR;

	/* ȼѡ֥å */
	superBlock[din]->blockSize=EXT2_BLOCK_BASE<<superBlock[din]->s_log_block_size;
	superBlock[din]->sectorSize=dev_stat.sect_size;
	superBlock[din]->sectors=superBlock[din]->blockSize/superBlock[din]->sectorSize;

	sectors=superBlock[din]->sectors;

	/* 롼ץǥץɤ߹ࡣ */
	a=ROUNDUP_DIV(superBlock[din]->s_blocks_count,superBlock[din]->s_blocks_per_group)*sizeof(GROUP_DESC);
	groupDescSize[din]=ROUNDUP_DIV(a,superBlock[din]->blockSize);
	if((groupDesc[din]=(GROUP_DESC*)kmalloc(superBlock[din]->blockSize*groupDescSize[din]))==NULL)goto ERR;
	if(read_direct(din,groupDesc[din],sectors*groupDescSize[din],blockToSect(din,EXT2_GROUP_DESC_BLK+superBlock[din]->s_first_data_block))!=
		sectors*groupDescSize[din])goto ERR;

	/* ֥åطν */
	if((gd=initBlock(din))==-1)goto ERR;

	/* inodeطν */
	if(initInode(din,gd)==-1)goto ERR;

	return (int)rootInode[din];

ERR:
	kfree(groupDesc[din]);
	groupDesc[din]=NULL;
	kfree(superBlock[din]);
	superBlock[din]=NULL;

	return -1;
}


static int umount(int din)
{
	umountInode(din);
	umountBlock(din);
	writeBackSbGd(din);

	kfree(groupDesc[din]);
	groupDesc[din]=NULL;

	kfree(superBlock[din]);
	superBlock[din]=NULL;

	return 0;
}


/*
 * parameters : path,device inode,directory inode address,truncate flag
 * return : inode buffer or error number
 */
static int open(const char *path,int din,uint dir_inode,int flag)
{
	int rest;
	INODE *inode,*inode_buf;
	INODE_PARAM param;


	if((param.inodeBuf=(INODE*)kmalloc(superBlock[din]->sectorSize))==NULL)return -ENOMEM;
	param.path=path;
	param.dirInode=(INODE*)dir_inode;
	if((rest=searchPath(din,&param))<0)goto ERR;
	inode=(INODE*)rest;
	if((inode->i_mode&EXT2_S_IFREG)==0)
	{
		if(inode->i_mode&EXT2_S_IFDIR)rest=-EISDIR;
		else rest=-ENOENT;
		goto ERR;
	}

	/* truncate. */
	if(flag!=0)
		if((rest=releaseDataBlock(din,inode))<0)goto ERR;

	if((inode_buf=(INODE*)kmalloc(sizeof(INODE)))!=NULL)
		memcpy(inode_buf,inode,sizeof(INODE));
	else (int)inode_buf=-ENOMEM;

	kfree(param.inodeBuf);

	return (int)inode_buf;
ERR:
	kfree(param.inodeBuf);

	return rest;
}


static int close(int din,uint inode)
{
	int rest;
	INODE *sector_buf;


	if((sector_buf=(INODE*)kmalloc(superBlock[din]->sectorSize))==NULL)return -ENOMEM;
	if((rest=readInode(din,((INODE*)inode)->i_number,sector_buf))<0)goto ERR;
	memcpy((void*)rest,(void*)inode,sizeof(INODE));
	if((rest=writeInode(din,((INODE*)inode)->i_number,sector_buf))<0)goto ERR;

	kfree(sector_buf);
	kfree((void*)inode);
	return 0;

ERR:
	kfree(sector_buf);
	kfree((void*)inode);
	return rest;
}


static int read(int din,uint inode,void *buf,size_t size,size_t begin)
{
	/* γǧ */
	if(begin>=((INODE*)inode)->i_size)return 0;
	if(begin+size>=((INODE*)inode)->i_size)size=((INODE*)inode)->i_size-begin;

	/* ե뤫ɤ߹ࡣ */
	if(readData(din,(INODE*)inode,buf,size,begin)!=size)return -1;

	/* դι */
	((INODE*)inode)->i_atime=(uint)sys_time();

	return size;
}


static int write(int din,uint inode,void *buf,size_t size,size_t begin)
{
	/* γǧ */
	if(begin>((INODE*)inode)->i_size)return -1;
	if(begin+size>((INODE*)inode)->i_size)
		if((begin+size-((INODE*)inode)->i_size)/superBlock[din]->blockSize>=superBlock[din]->s_free_blocks_count)
			return -1;

	/* եؽ񤭹ߡ */
	if(writeData(din,(INODE*)inode,buf,size,begin)!=size)return -1;


	/* ι */
	if(begin+size>((INODE*)inode)->i_size)
	{
		((INODE*)inode)->i_size=begin+size;
		((INODE*)inode)->i_blocks=ROUNDUP_DIV(begin+size,INODE_BLOCK_SIZE);
	}

	/* դι */
	((INODE*)inode)->i_atime=((INODE*)inode)->i_mtime=(uint)sys_time();

	return size;
}


static int fstat(uint inode,FSTAT *fstat)
{
	fstat->st_mode=((INODE*)inode)->i_mode;
	fstat->st_nlink=((INODE*)inode)->i_links_count;
	fstat->st_uid=((INODE*)inode)->i_uid;
	fstat->st_gid=((INODE*)inode)->i_gid;
	fstat->st_size=((INODE*)inode)->i_size;
	fstat->st_atime=((INODE*)inode)->i_atime;
	fstat->st_mtime=((INODE*)inode)->i_mtime;
	fstat->st_ctime=((INODE*)inode)->i_ctime;

	return 0;
}


static int mkdir(const char *path,int din,uint dir_inode,int mode)
{
	uint inumber;
	uint block;
	int rest;
	INODE *inode;
	INODE_PARAM iparam;
	PROC *proc;


	/* ѥθ */
	if((iparam.inodeBuf=(INODE*)kmalloc(superBlock[din]->sectorSize))==NULL)return -ENOMEM;
	iparam.path=path;
	iparam.dirInode=(INODE*)dir_inode;
	rest=searchPath(din,&iparam);
	if(rest>0)
	{
		rest=-EEXIST;
		goto ERR;
	}
	else if((rest!=-ENOENT)||((rest=isLastPath(iparam.path))==-1))goto ERR;

	/* ե̾Υå */
	if(rest>EXT2_NAME_LEN)
	{
		rest=-ENAMETOOLONG;
		goto ERR;
	}

	/* inode롣 */
	if((inumber=getInodeNumber(din))==0)
	{
		rest=-ENOSPC;
		goto ERR;
	}

	/* ǥ쥯ȥꥨȥ꡼ɲä롣 */
	if((rest=addDirEnt(din,EXT2_FT_DIR,inumber,&iparam))!=0)goto ERR;

	/* ƥǥ쥯ȥΥ󥯿1䤹 */
	++iparam.dirInode->i_links_count;
	if((uint)iparam.dirInode!=dir_inode)
		if((rest=writeInode(din,iparam.dirInode->i_number,iparam.inodeBuf))!=0)goto ERR;

	/* ǥ쥯ȥ롣 */
	if((block=getBlock(din))==0)
	{
		rest=-ENOSPC;
		goto ERR;
	}
	if((rest=initDirBlock(din,block,inumber,iparam.dirInode->i_number))!=0)goto ERR;

	/* inode롣 */
	if(((int)inode=readInode(din,inumber,iparam.inodeBuf))<0)
	{
		rest=(int)inode;
		goto ERR;
	}
	memset(inode,0,sizeof(INODE));
	proc=get_current_task();
	inode->i_mode=mode|EXT2_S_IFDIR;
	inode->i_size=superBlock[din]->blockSize;
	inode->i_uid=proc->uid;
	inode->i_gid=proc->gid;
	inode->i_atime=inode->i_ctime=inode->i_mtime=(uint)sys_time();
	inode->i_links_count=2;				/* "."".."Σġ */
	inode->i_blocks=inode->i_size/INODE_BLOCK_SIZE;
	inode->i_block[0]=block;
	if((rest=writeInode(din,inumber,iparam.inodeBuf))<0)goto ERR;

	kfree(iparam.inodeBuf);
	return 0;

ERR:
	kfree(iparam.inodeBuf);
	return rest;
}


/*
 * return : inode buffer address or error number
 */
static int creat(const char *path,int din,uint dir_inode,int mode)
{
	uint inumber;
	int rest;
	INODE *inode,*inode_buf;
	INODE_PARAM iparam;
	PROC *proc;


	/* ѥθ */
	if((iparam.inodeBuf=(INODE*)kmalloc(superBlock[din]->sectorSize))==NULL)return -ENOMEM;
	iparam.path=path;
	iparam.dirInode=(INODE*)dir_inode;
	rest=searchPath(din,&iparam);
	if(rest>0)
	{
		rest=-EEXIST;
		goto ERR;
	}
	else if((rest!=-ENOENT)||((rest=isLastPath(iparam.path))==-1))goto ERR;

	/* ե̾Υå */
	if(rest>EXT2_NAME_LEN)
	{
		rest=-ENAMETOOLONG;
		goto ERR;
	}

	/* inode롣 */
	if((inumber=getInodeNumber(din))==0)
	{
		rest=-ENOSPC;
		goto ERR;
	}

	/* ǥ쥯ȥꥨȥ꡼ɲä롣 */
	if((rest=addDirEnt(din,EXT2_FT_REG_FILE,inumber,&iparam))!=0)goto ERR;

	/* inode롣 */
	if(((int)inode=readInode(din,inumber,iparam.inodeBuf))<0)
	{
		rest=(int)inode;
		goto ERR;
	}
	memset(inode,0,sizeof(INODE));
	proc=get_current_task();
	inode->i_mode=mode|EXT2_S_IFREG;
	inode->i_uid=proc->uid;
	inode->i_gid=proc->gid;
	inode->i_atime=inode->i_ctime=inode->i_mtime=(uint)sys_time();
	inode->i_links_count=1;
	inode->i_number=inumber;
	if((rest=writeInode(din,inumber,iparam.inodeBuf))<0)goto ERR;

	/* inodeinodeХåե˥ԡ롣 */
	if((inode_buf=(INODE*)kmalloc(sizeof(INODE)))==NULL)rest=-ENOMEM;
	else memcpy(inode_buf,inode,sizeof(INODE));

	kfree(iparam.inodeBuf);
	return (int)inode_buf;

ERR:
	kfree(iparam.inodeBuf);
	return rest;
}


static int delete(const char *path,int din,uint dir_inode,int type)
{
	int rest;
	INODE *inode;
	INODE_PARAM param;


	/* ƥǥ쥯ȥ꤫ǥ쥯ȥꥨȥ꡼롣 */
	if((param.inodeBuf=(INODE*)kmalloc(superBlock[din]->sectorSize))==NULL)return -ENOMEM;
	param.path=path;
	param.dirInode=(INODE*)dir_inode;
	if(type==NORMAL_FILE)
	{
		if((rest=delFileEntry(din,&param))<0)goto END;
	}
	else
	{
		if((rest=delDirEntry(din,&param))<0)goto END;
	}
	inode=(INODE*)rest;

	/* ǡ֥å롣 */
	if((rest=releaseDataBlock(din,inode))<0)goto END;

	/* inode롣 */
	releaseInode(din,inode->i_number);

	rest=0;
END:
	kfree(param.inodeBuf);

	return rest;
}


/*********************************** ̤ ***************************************/
static int opendir(const char *path,int din,uint dir_inode)
{
	return 0;
}


static int readdir(int din,uint inode,struct dirent *dent)
{
	return 0;
}


static int ioctr(int din,uint inode,int cmd,void *param)
{
	return 0;
}


static int rename(int din,uint inode,const char *name)
{
	return 0;
}


/***********************************************************************
 *
 * < Init class >
 *
 ***********************************************************************/

static FS ext2Fs={
	"ext2",mount,umount,open,close,read,write,ioctr,rename,creat,opendir,mkdir,fstat,delete,readdir
};


/*
 * GLOBAL
 */
int initExt2tFs()
{
	int i;


	for(i=0;i<MAX_DEVICE_OPEN;++i)
	{
		rootInode[i]=NULL;
		groupDesc[i]=NULL;
		superBlock[i]=NULL;
	}

	return regist_fs(&ext2Fs);
}
/**************************************************************************************/
int test_ext2()
{
	int din;


	for(din=0;superBlock[din]==NULL;++din);
	umountInode(din);
	umountBlock(din);
	writeBackSbGd(din);

	return 0;
}
/**************************************************************************************/
